package tomo

import "image"
import "image/color"
import "git.tebibyte.media/tomo/tomo/canvas"

// Attr modifies one thing about a box's style.
type Attr interface {
	// Equals returns true if both attributes can reasonably be declared
	// equal.
	Equals (Attr) bool
	Kind   () AttrKind
	attr   ()
}

// AttrKind enumerates all attribute kinds. There is one of these for each
// attribute type.
type AttrKind int; const (
	AttrKindColor AttrKind = iota
	AttrKindTexture
	AttrKindTextureMode
	AttrKindBorder
	AttrKindMinimumSize
	AttrKindPadding
	AttrKindGap
	AttrKindTextColor
	AttrKindDotColor
	AttrKindFace
	AttrKindWrap
	AttrKindAlign
	AttrKindOverflow
	AttrKindLayout
	AttrKindCursor
)

// AttrColor sets the background color of a box.
type AttrColor struct { color.Color }
// AttrTexture sets the texture of a box to a texture.
type AttrTexture struct { canvas.Texture }
// AttrTextureMode sets the rendering mode of a box's texture.
type AttrTextureMode TextureMode
// AttrBorder sets the border of a box.
type AttrBorder []Border
// AttrMinimumSize sets the minimum size of a box.
type AttrMinimumSize image.Point
// AttrPadding sets the inner padding of a box.
type AttrPadding Inset
// AttrGap sets the gap between child boxes, if the box is a ContainerBox.
type AttrGap image.Point
// AttrTextColor sets the text color, if the box is a TextBox.
type AttrTextColor struct { color.Color }
// AttrDotColor sets the text selection color, if the box is a TextBox.
type AttrDotColor struct { color.Color }
// AttrFace sets the type face, if the box is a TextBox.
type AttrFace Face
// AttrWrap sets if the text wraps, if the box is a TextBox.
type AttrWrap bool
// AttrAlign sets the alignment, if the box is a ContentBox.
type AttrAlign struct { X, Y Align }
// AttrOverflow sets the overflow, if the box is a ContentBox.
type AttrOverflow struct { X, Y bool }
// AttrLayout sets the layout, if the box is a ContentBox.
type AttrLayout struct { Layout }
// AttrCursor sets the mouse cursor shape.
type AttrCursor Cursor

// AColor is a convenience constructor for the color attribute.
func AColor (col color.Color) AttrColor {
	return AttrColor { Color: col }
}
// ATexture is a convenience constructor for the texture attribute.
func ATexture (texture canvas.Texture) AttrTexture {
	return AttrTexture { Texture: texture }
}
// ATextureMode is a convenience constructor for the texture mode attribute.
func ATextureMode (mode TextureMode) AttrTextureMode {
	return AttrTextureMode(mode)
}
// ABorder is a convenience constructor for the border attribute.
func ABorder (borders ...Border) AttrBorder {
	return AttrBorder(borders)
}
// AMinimumSize is a convenience constructor for the minimum size attribute.
func AMinimumSize (x, y int) AttrMinimumSize {
	return AttrMinimumSize(image.Pt(x, y))
}
// APadding is a convenience constructor for the padding attribute.
func APadding (sides ...int) AttrPadding {
	return AttrPadding(I(sides...))
}
// AGap is a convenience constructor for the gap attribute.
func AGap (x, y int) AttrGap {
	return AttrGap(image.Pt(x, y))
}
// ATextColor is a convenience constructor for the text color attribute.
func ATextColor (col color.Color) AttrTextColor {
	return AttrTextColor { Color: col }
}
// ADotColor is a convenience constructor for the dot color attribute.
func ADotColor (col color.Color) AttrDotColor {
	return AttrDotColor { Color: col }
}
// AFace is a convenience constructor for the face attribute.
func AFace (face Face) AttrFace {
	return AttrFace(face)
}
// AWrap is a convenience constructor for the wrap attribute.
func AWrap (wrap bool) AttrWrap {
	return AttrWrap(wrap)
}
// AAlign is a convenience constructor for the align attribute.
func AAlign (x, y Align) AttrAlign {
	return AttrAlign { X: x, Y: y }
}
// AOverflow is a convenience constructor for the overflow attribute.
func AOverflow (x, y bool) AttrOverflow {
	return AttrOverflow { X: x, Y: y }
}
// ALayout is a convenience constructor for the overflow attribute.
func ALayout (layout Layout) AttrLayout {
	return AttrLayout { Layout: layout }
}
// ACursor is a convenience constructor for the cursor attribute.
func ACursor (cursor Cursor) AttrCursor {
	return AttrCursor(cursor)
}

// Equals returns true if both attributes can reasonably be declared equal.
func (this AttrColor) Equals (other Attr) bool {
	if other, ok := other.(AttrColor); ok {
		return this == other
	} else {
		return false
	}
}
// Equals returns true if both attributes can reasonably be declared equal.
func (this AttrTexture) Equals (other Attr) bool {
	if other, ok := other.(AttrTexture); ok {
		return this == other
	} else {
		return false
	}
}
// Equals returns true if both attributes can reasonably be declared equal.
func (this AttrTextureMode) Equals (other Attr) bool {
	if other, ok := other.(AttrTextureMode); ok {
		return this == other
	} else {
		return false
	}
}
// Equals returns true if both attributes can reasonably be declared equal.
func (this AttrBorder) Equals (other Attr) bool {
	if other, ok := other.(AttrBorder); ok {
		if len(this) != len(other) { return false }
		for index := range this {
			thisBorder  := this[index]
			otherBorder := other[index]
			if thisBorder != otherBorder { return false }
		}
		return true
	} else {
		return false
	}
}
// Equals returns true if both attributes can reasonably be declared equal.
func (this AttrMinimumSize) Equals (other Attr) bool {
	if other, ok := other.(AttrMinimumSize); ok {
		return this == other
	} else {
		return false
	}
}
// Equals returns true if both attributes can reasonably be declared equal.
func (this AttrPadding) Equals (other Attr) bool {
	if other, ok := other.(AttrPadding); ok {
		return this == other
	} else {
		return false
	}
}
// Equals returns true if both attributes can reasonably be declared equal.
func (this AttrGap) Equals (other Attr) bool {
	if other, ok := other.(AttrGap); ok {
		return this == other
	} else {
		return false
	}
}
// Equals returns true if both attributes can reasonably be declared equal.
func (this AttrTextColor) Equals (other Attr) bool {
	if other, ok := other.(AttrTextColor); ok {
		return this == other
	} else {
		return false
	}
}
// Equals returns true if both attributes can reasonably be declared equal.
func (this AttrDotColor) Equals (other Attr) bool {
	if other, ok := other.(AttrDotColor); ok {
		return this == other
	} else {
		return false
	}
}
// Equals returns true if both attributes can reasonably be declared equal.
func (this AttrFace) Equals (other Attr) bool {
	if other, ok := other.(AttrFace); ok {
		return this == other
	} else {
		return false
	}
}
// Equals returns true if both attributes can reasonably be declared equal.
func (this AttrWrap) Equals (other Attr) bool {
	if other, ok := other.(AttrWrap); ok {
		return this == other
	} else {
		return false
	}
}
// Equals returns true if both attributes can reasonably be declared equal.
func (this AttrAlign) Equals (other Attr) bool {
	if other, ok := other.(AttrAlign); ok {
		return this == other
	} else {
		return false
	}
}
// Equals returns true if both attributes can reasonably be declared equal.
func (this AttrOverflow) Equals (other Attr) bool {
	if other, ok := other.(AttrOverflow); ok {
		return this == other
	} else {
		return false
	}
}
// Equals returns true if both attributes can reasonably be declared equal.
func (this AttrLayout) Equals (other Attr) bool {
	// for some goofy reason if we try to compare two AttrLayouts we get a
	// fucking runtime error????? its probably for the best anyways because
	// two layouts cannot "reasonably" be declared equal
	return false
}
// Equals returns true if both attributes can reasonably be declared equal.
func (this AttrCursor) Equals (other Attr) bool {
	if other, ok := other.(AttrCursor); ok {
		return this == other
	} else {
		return false
	}
}

func (AttrColor)       Kind () AttrKind { return AttrKindColor }
func (AttrTexture)     Kind () AttrKind { return AttrKindTexture }
func (AttrTextureMode) Kind () AttrKind { return AttrKindTextureMode }
func (AttrBorder)      Kind () AttrKind { return AttrKindBorder }
func (AttrMinimumSize) Kind () AttrKind { return AttrKindMinimumSize }
func (AttrPadding)     Kind () AttrKind { return AttrKindPadding }
func (AttrGap)         Kind () AttrKind { return AttrKindGap }
func (AttrTextColor)   Kind () AttrKind { return AttrKindTextColor }
func (AttrDotColor)    Kind () AttrKind { return AttrKindDotColor }
func (AttrFace)        Kind () AttrKind { return AttrKindFace }
func (AttrWrap)        Kind () AttrKind { return AttrKindWrap }
func (AttrAlign)       Kind () AttrKind { return AttrKindAlign }
func (AttrOverflow)    Kind () AttrKind { return AttrKindOverflow }
func (AttrLayout)      Kind () AttrKind { return AttrKindLayout }
func (AttrCursor)      Kind () AttrKind { return AttrKindCursor }

func (AttrColor)       attr () { }
func (AttrTexture)     attr () { }
func (AttrTextureMode) attr () { }
func (AttrBorder)      attr () { }
func (AttrMinimumSize) attr () { }
func (AttrPadding)     attr () { }
func (AttrGap)         attr () { }
func (AttrTextColor)   attr () { }
func (AttrDotColor)    attr () { }
func (AttrFace)        attr () { }
func (AttrWrap)        attr () { }
func (AttrAlign)       attr () { }
func (AttrOverflow)    attr () { }
func (AttrLayout)      attr () { }
func (AttrCursor)      attr () { }
