// Package event provides a system for broadcasting events to multiple event
// handlers.
package event

import "io"
import "errors"

// A cookie is returned when you add an event handler so you can remove it
// later if you so choose. When the Close behavior is called, the handler must
// be removed, even if an error is returned.
type Cookie io.Closer

// FuncCookie is a cookie that calls a function (itself) when closed.
type FuncCookie func () error
func (cookie FuncCookie) Close () error { return cookie () }

// NoCookie is a cookie that does nothing when closed.
type NoCookie struct { }
func (NoCookie) Close () error { return nil }

type multiCookie []Cookie

// MultiCookie creates a single cookie that, when closed, closes a list of other
// cookies.
func MultiCookie (cookies ...Cookie) Cookie {
	return multiCookie(cookies)
}

func (cookies multiCookie) Close () error {
	errs := make([]error, len(cookies))
	for index, cookie := range cookies {
		errs[index] = cookie.Close()
	}
	return errors.Join(errs...)
}

// Broadcaster manages event listeners.
type Broadcaster[L any] struct {
	lastID    int
	listeners map[int] L
}

// Connect adds a new listener to the broadcaster and returns a corresponding
// cookie.
func (broadcaster *Broadcaster[L]) Connect (listener L) Cookie {
	broadcaster.ensure()
	
	cookie := broadcaster.newCookie()
	broadcaster.listeners[cookie.id] = listener
	return cookie
}

// Listeners returns a map of all connected listeners.
func (broadcaster *Broadcaster[L]) Listeners () map[int] L {
	broadcaster.ensure()
	return broadcaster.listeners
}

func (broadcaster *Broadcaster[L]) newCookie () broadcasterCookie[L] {
	broadcaster.lastID ++
	return broadcasterCookie[L] {
		id:          broadcaster.lastID,
		broadcaster: broadcaster,
	}
}

func (broadcaster *Broadcaster[L]) ensure () {
	if broadcaster.listeners == nil {
		broadcaster.listeners = make(map[int] L)
	}
}

type broadcasterCookie[L any] struct {
	id int
	broadcaster *Broadcaster[L]
}

func (cookie broadcasterCookie[L]) Close () error {
	delete(cookie.broadcaster.listeners, cookie.id)
	return nil
}

// FuncBroadcaster is a broadcaster that manages functions with no arguments.
type FuncBroadcaster struct {
	Broadcaster[func ()]
}

// Broadcast calls all connected listener funcs.
func (broadcaster *FuncBroadcaster) Broadcast () {
	for _, listener := range broadcaster.Listeners() {
		listener()
	}
}
